#lang scheme

#|
  <expr> = <number>
         | <id>
         | (+ <expr> <expr>)
         | (- <expr> <expr>)
         | (zero? <expr>)
         | (if <expr> <expr> <expr>)
         | (let ([<id> <expr>]) <expr>)
         | (lambda (<id>) <expr>)
         | (<expr> <expr>)
|#

(define-struct closure (arg body env))

(define (interp expr env)
  (cond
    [(number? expr) expr]
    [(symbol? expr)
     (cdr (assq expr env))]
    [else
     (case (car expr)
       [(+)
        (+ (interp (cadr expr) env)
           (interp (caddr expr) env))]
       [(-)
        (- (interp (cadr expr) env)
           (interp (caddr expr) env))]
       [(zero?)
        (zero? (interp (cadr expr) env))]
       [(if)
        (if (interp (cadr expr) env)
            (interp (caddr expr) env)
            (interp (cadddr expr) env))]
       [(let)
        (let ([rhs-val
               (interp (cadr (caadr expr)) env)])
          (interp (caddr expr)
                  (cons (cons (car (caadr expr))
                              rhs-val)
                        env)))]
       [(lambda)
        (make-closure (caadr expr)
                      (caddr expr)
                      env)]
       [else
        (let ([clos (interp (car expr) env)]
              [arg-val (interp (cadr expr) env)])
          (interp (closure-body clos)
                (cons
                 (cons (closure-arg clos)
                       arg-val)
                 (closure-env clos))))])]))

(interp '(let ([f (lambda (x) 
                    (lambda (y)
                      (- y x)))]) 
           (+ ((f 10) 2) 3))
        '())


(time (interp '(let ([Y (lambda (m)
                          ((lambda (f) (m (lambda (a) ((f f) a))))
                           (lambda (f) (m (lambda (a) ((f f) a))))))])
                 (let ([sum
                        (Y (lambda (sum)
                             (lambda (n)
                               (if (zero? n)
                                   0
                                   (+ n (sum (- n 1)))))))])
                   (sum 50000)))
              '()))

(time (let ([Y (lambda (m)
                 ((lambda (f) (m (lambda (a) ((f f) a))))
                  (lambda (f) (m (lambda (a) ((f f) a))))))])
        (let ([sum
               (Y (lambda (sum)
                    (lambda (n)
                      (if (zero? n)
                          0
                          (+ n (sum (- n 1)))))))])
          (sum 50000))))

(time (letrec ([sum
                (lambda (n)
                  (if (zero? n)
                      0
                      (+ n (sum (- n 1)))))])
        (sum 50000)))
